
UNDEFINED = -1


RS_TYPES_KNOWN = 1
RS_SLICES_KNOWN = 2
RS_ALL_KNOWN = RS_TYPES_KNOWN | RS_SLICES_KNOWN
RS_INFERRED = 8
RS_CHECK = 16


BCCOPY=0
BCENSURE=1
BCEXIST=2
BCNEW=3


NOVAL="NOVALUE" #alternative None for functions which also accept NOne as valid input
                #but still want some params to be optional
